/*----------------------------------------------------------------------
TitanBomber - Roberto Prieto
Copyright (C) 2005-2023 MegaStorm Systems
contact@megastormsystems.com - http://www.megastormsystems.com

This software is provided 'as-is', without any express or implied
warranty.  In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must not
claim that you wrote the original software. If you use this software
in a product, an acknowledgment in the product documentation would be
appreciated but is not required.
2. Altered source versions must be plainly marked as such, and must not be
misrepresented as being the original software.
3. This notice may not be removed or altered from any source distribution.

------------------------------------------------------------------------

gameField class

------------------------------------------------------------------------ */

#include "gamefield.h"
#include "objects.h"
#include "resource.h"

// Pointer to instanced gamefield used by myCoreServer static function
cGameField *myGameField = NULL;

// Wrapper for the Render Graphics callback
void myRenderInGame(int i)
{
    myGameField->RenderGraphics(i);
}

// Constructor
cGameField::cGameField()
{
    int i,j,k;

    // Debugging system
    bDebug = -1;
    bMapReady = 0;
    iDebugUpdate = 0; Type = 255;
    fp1 = fp2 = fp3 = fp4 = NULL;

    // Misc
    iNumPlayers = iNumRemotePlayers = 0;
    mutMap = NULL; fFPS = 0;
    strcpy(host,"localhost");
    iTimeStart = iTimeEnd = iTimeMatch = 0;
    strcpy(sCounterDown,"");
    vFloatingMsgs.clear();
    strcpy(sMsg,"");
    strcpy(sTmp,"");

    // Reset map list
    memset(&lstMapGame,-1,sizeof(lstMapGame));

    // Mapa 1: Water field
    strcpy(lstMapGame[0].sName,"Water");
    lstMapGame[0].iOwner = PLAYER_WATER;
    lstMapGame[0].iNumBK = 90; 
    // Mapa 2: Fire field
    strcpy(lstMapGame[1].sName,"Fire");
    lstMapGame[1].iOwner = PLAYER_FIRE;
    lstMapGame[1].iNumBK = 90; 
    // Mapa 3: Earth field
    strcpy(lstMapGame[2].sName,"Earth");
    lstMapGame[2].iOwner = PLAYER_EARTH;
    lstMapGame[2].iNumBK = 90; 
    // Mapa 4: Air field
    strcpy(lstMapGame[3].sName,"Air");
    lstMapGame[3].iOwner = PLAYER_AIR;
    lstMapGame[3].iNumBK = 90; 

    // Prepare all maps: resource names, status, etc.
    for(i = 0; i < NUMBER_OF_MAP; i++)
    {
        strcpy(lstMapGame[i].sBg,lstMapGame[i].sName);
        strcat(lstMapGame[i].sBg,"_background");

        strcpy(lstMapGame[i].sBlock,lstMapGame[i].sName);
        strcat(lstMapGame[i].sBlock,"_block"); 

        strcpy(lstMapGame[i].sBreakable,lstMapGame[i].sName);
        strcat(lstMapGame[i].sBreakable,"_breakable");

        memset(lstMapGame[i].tMapBomb,0,sizeof(tMapGame.tMapBomb));
        for(k = 0; k < MAP_WIDTH; k++)
            for(j = 0; j < MAP_HEIGHT; j++)
            {				
                lstMapGame[i].tMapBomb[k][j].status = STATUS_WALKABLE;
            }	
            for(k = 1; k < MAP_WIDTH; k = k + 2)
                for(j = 1; j < MAP_HEIGHT; j = j + 2)
                {
                    lstMapGame[i].tMapBomb[k][j].item = ITEM_BLOCK;
                    lstMapGame[i].tMapBomb[k][j].status = STATUS_NOWALKABLE;
                }	
    }

    // Load background for end of match screen
    sEndMatchBg = IImage->Load(GFX_RESOURCE,"EndMatch_bg");

    // External pointer to this instance
    myGameField = this;

    ILogSystem.Msg(LOG_NORMAL,"\n [GameField Info] - Initialized.\n");
}

// Destructor
cGameField::~cGameField()
{
    if(fp1 != NULL) fclose(fp1);
    if(fp2 != NULL) fclose(fp2);
    if(fp3 != NULL) fclose(fp3);
    if(fp4 != NULL) fclose(fp4);
    fp1 = fp2 = fp3 = fp4 = NULL;
    memset(&lstMapGame,-1,sizeof(lstMapGame));
    vFloatingMsgs.clear();

    // Free background for end of match screen
    CRM32Pro.FreeSurface(sEndMatchBg);

    // External pointer to this instance
    myGameField = NULL;

    ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Destroyed.\n\n");
}

// Close current match but leaves it ready to init another one
void cGameField::Close()
{
    Uint32 i;

    // Network match, kill server and clients
#ifdef DEBUG_NETWORK
    if(Mode != GAMEMODE_LOCAL) INetwork->Info();
#endif
    if(Mode == GAMEMODE_SERVER) INetwork->QueryKillServer();
    if(Mode == GAMEMODE_CLIENT) INetwork->QueryKillClient();
    if(mutMap) SDL_DestroyMutex(mutMap);
    mutMap = NULL;
    // Network matches: wait 2seg before closing the network but keep updating the screen, events, etc.
    if(Mode != GAMEMODE_LOCAL)
    {
        i = SDL_GetTicks();
        while( (i + 2000) > SDL_GetTicks())
        {
            while(CRM32Pro.Update());
            SDL_Delay(10); 
        }			
        INetwork->Quit();

        // Inmediately close the console (if it is still visible)
        CRM32Pro.ConFeatures(CONFEATURE_DISABLE_FADE); // enable this flag
        CRM32Pro.ConSetVisible(0);
        CRM32Pro.ConFeatures(CONFEATURE_DISABLE_FADE); // disable this flag
        CRM32Pro.ConCmd("con clear");
    }

    // Show gamefield information
    if(bMapReady) 
        ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Map '%s' execution at %2.2f FPS during %2.2fSec is over\n",	GetMapName(),fFPS,(float)(iTimeEnd - iTimeStart) / 1000.0f);

    // Reset vars
    bMapReady = 0;
    iDebugUpdate = 0; Type = 255;
    iTimeStart = iTimeEnd = iTimeMatch = 0;
    strcpy(sCounterDown,"");
    iNumPlayers = iNumRemotePlayers = 0;
    vFloatingMsgs.clear();
    for(i = 0; i < 4; i++)
    {
        if(!TBDN_Events[i].empty())
        {
            // We never should reach this point...but just for debugging purposes
            ILogSystem.Msg(LOG_NORMAL," [GameField Warning] - P%d still had events(%d) for processing...\n",i,TBDN_Events[i].size());
            while(!TBDN_Events[i].empty()) TBDN_Events[i].pop(); // clear the queue
        }
    }
    ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Game field closed.\n\n");
}

// Add a new floating message. 
void cGameField::AddFloatingMsg(int x,int y,char *payload,CRM32Pro_CFont *font,int ttl)
{
    stFloatingMsg stMyMsg;

    // Checks some parameters
    if(payload == NULL) return;
    if(ttl < 0) return;
    if(font == NULL) return;

    // Create the message and push back into the list
    stMyMsg.x = x; stMyMsg.y = y; 
    stMyMsg.iTTL = ttl;
    stMyMsg.iStart = ITimeSystem->GetTime();
    stMyMsg.fMsg = font;
    strcpy(stMyMsg.sPayload,payload);
    vFloatingMsgs.push_back(stMyMsg);
    //ILogSystem.Msg(LOG_NORMAL," > New floating message (%d,%d): '%s' font '%x'\n",x,y,payload,font);
}

// Return the played time of the previous match (in seconds)
int cGameField::GetPlayedTime()
{
    return (iTimeEnd - iTimeStart) / 1000;
}

// Return current loaded map name
char *cGameField::GetMapName()
{
    if(bMapReady) return tMapGame.sName;
    else return "Map not found";
}

// Return map name of the given position
char *cGameField::GetMapName(int nMap)
{
    if(nMap < 0) return NULL;
    if(nMap > NUMBER_OF_MAP) return NULL;
    return lstMapGame[nMap].sName;
}

// Return the gametype name
char *cGameField::GetGameTypeName()
{
    switch(Type)
    {
    case GAMETYPE_LEAGUE_LAST :
    case GAMETYPE_LEAGUE:
        return "League";
        break;
    case GAMETYPE_TIMED:
        return "Timed";
        break;
    case GAMETYPE_RANDOM:
        return "Random";
        break;
    }
    return NULL;
}

// Return the gametype name
int cGameField::GetGameType()
{
    return Type;
}

// Load a map and perform a random generation of breakables+items
void cGameField::MapLoad(char *fmap)
{
    int i,k,j,iNumPasadas=0;
    int iNumBK=0,iNumF=0,iNumB=0,iNumSU=0,iNumSD=0;

    // Look for the map
    for(i=0;i<NUMBER_OF_MAP;i++)
    {
        if(!strcmp(fmap,lstMapGame[i].sName)) 
        { 
            strcpy(tMapGame.sBg,lstMapGame[i].sBg); 
            strcpy(tMapGame.sBlock,lstMapGame[i].sBlock); 
            strcpy(tMapGame.sBreakable,lstMapGame[i].sBreakable); 
            strcpy(tMapGame.sName,fmap); 
            tMapGame.iNumBK=lstMapGame[i].iNumBK;
            // Max adjust
            if(tMapGame.iNumBK>MAX_BREAKABLES) tMapGame.iNumBK=MAX_BREAKABLES;
            tMapGame.iOwner=lstMapGame[i].iOwner;
            if(tMapGame.iNumBK<(ITEMS_BOMBS+ITEMS_FIREUP+ITEMS_SPEEDUP+ITEMS_SPEEDDOWN)) tMapGame.iNumBK=ITEMS_BOMBS+ITEMS_FIREUP+ITEMS_SPEEDUP+ITEMS_SPEEDDOWN;
            memcpy(tMapGame.tMapBomb,lstMapGame[i].tMapBomb,sizeof(tMapGame.tMapBomb));	
            // Random generation of breakables+items	
            while(iNumBK!=tMapGame.iNumBK || iNumB<ITEMS_BOMBS || iNumF<ITEMS_FIREUP || iNumSU<ITEMS_SPEEDUP || iNumSD<ITEMS_SPEEDDOWN)
            {
                for(k = 0; k < MAP_WIDTH; k++)
                    for(j = 0; j < MAP_HEIGHT; j++) 
                    {
                        // 4 corners are free
                        if(k == 0) if(j == 0 || j == 1 || j == MAP_HEIGHT - 2 || j == MAP_HEIGHT - 1) continue;
                        if(k == 1) if(j == 0 || j == MAP_HEIGHT - 1) continue;
                        if(k == MAP_WIDTH - 1) if(j == 0 || j == 1 || j == MAP_HEIGHT - 2 || j == MAP_HEIGHT - 1) continue;
                        if(k == MAP_WIDTH - 2) if(j == 0 || j == MAP_HEIGHT - 1) continue;

                        // Si no hay un bloque fijo y tampoco hay otro breakable, colocamos uno con el 20% de probabilidad
                        if(GetItem(k,j)==ITEM_NULL && iNumBK!=tMapGame.iNumBK)
                        {							
                            if((CRM32Pro.Rand()%100)<=20)
                            {
                                tMapGame.tMapBomb[k][j].item=ITEM_BREAKABLE;
                                tMapGame.tMapBomb[k][j].status=STATUS_NOWALKABLE;
                                iNumBK++;
                            }
                        }

                        // Si tenemos un breakable ya instalado, con probabilidad del 5% colocamos un item bomba
                        if(GetItem(k,j)==ITEM_BREAKABLE && iNumB<ITEMS_BOMBS)
                        {
                            if((CRM32Pro.Rand()%100)<=5)
                            {
                                tMapGame.tMapBomb[k][j].item=ITEM_BK_BOMB;
                                iNumB++;
                            }
                        }

                        // Si tenemos un breakable ya instalado, con probabilidad del 5% colocamos un item aumento explosion
                        if(GetItem(k,j)==ITEM_BREAKABLE && iNumF<ITEMS_FIREUP)
                        {
                            if((CRM32Pro.Rand()%100)<=5)
                            {
                                tMapGame.tMapBomb[k][j].item=ITEM_BK_FIREUP;
                                iNumF++;
                            }
                        }

                        // Si tenemos un breakable ya instalado, con probabilidad del 5% colocamos un item aumento velocidad
                        if(GetItem(k,j)==ITEM_BREAKABLE && iNumSU<ITEMS_SPEEDUP)
                        {
                            if((CRM32Pro.Rand()%100)<=5)
                            {
                                tMapGame.tMapBomb[k][j].item=ITEM_BK_SPEEDUP; 
                                iNumSU++;
                            }
                        }

                        // Si tenemos un breakable ya instalado, con probabilidad del 5% colocamos un item descenso velocidad
                        if(GetItem(k,j)==ITEM_BREAKABLE && iNumSD<ITEMS_SPEEDDOWN)
                        {
                            if((CRM32Pro.Rand()%100)<=5)
                            {
                                tMapGame.tMapBomb[k][j].item=ITEMS_SPEEDDOWN; 
                                iNumSD++;
                            }
                        }
                    }
                    iNumPasadas++;
            }
            // Ya encontramos y generamos el mapa
            break; 
        }
    }
    if(i == NUMBER_OF_MAP) { bMapReady = 0; ILogSystem.Msg(LOG_NORMAL," [GameField Error] - Map '%s' not found.\n",fmap); }
    else { bMapReady = 1; ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Map '%s' loaded. Breakables: %d. Loop random: %d\n",fmap,tMapGame.iNumBK,iNumPasadas); }
}


// Get .status from the map given tx,ty
Uint8 cGameField::GetStatus(int tx,int ty)
{
    Uint8 tmp;

    // Check boundaries
    if(tx < 0) return STATUS_NULL;
    else if(tx > MAP_WIDTH - 1) return STATUS_NULL;
    if(ty < 0) return STATUS_NULL;
    else if(ty > MAP_HEIGHT - 1) return STATUS_NULL;

    // Get status
    if(Mode == GAMEMODE_SERVER) SDL_LockMutex(mutMap);
    tmp = tMapGame.tMapBomb[tx][ty].status;
    if(Mode == GAMEMODE_SERVER) SDL_UnlockMutex(mutMap);

    // Return
    return tmp;
}

// Set .status on the map given tx,ty and the value
// When it sets bomb explosion, it has an adding effect as more than one bomb can be affecting this cell
void cGameField::SetStatus(int tx,int ty,Uint8 status)
{
    // Check boundaries
    if(tx < 0) return;
    else if(tx > MAP_WIDTH - 1) return;
    if(ty < 0) return;
    else if(ty > MAP_HEIGHT - 1) return;

    if(Mode == GAMEMODE_SERVER) SDL_LockMutex(mutMap);
    // Is it a bomb explosion?
    // Add to previous state
    if(status >= STATUS_BURN_P1)
    {
        // Remove the bomb itself
        if(tMapGame.tMapBomb[tx][ty].status == STATUS_NOWALKABLE) tMapGame.tMapBomb[tx][ty].status = status; 
        else tMapGame.tMapBomb[tx][ty].status = tMapGame.tMapBomb[tx][ty].status | status;
    }
    // No, it is another state
    else tMapGame.tMapBomb[tx][ty].status = status;
    if(Mode == GAMEMODE_SERVER) SDL_UnlockMutex(mutMap);
}

// Get .item from the map given tx,ty
Uint8 cGameField::GetItem(int tx,int ty)
{
    Uint8 tmp;

    // Check boundaries
    if(tx < 0) return ITEM_NULL;
    else if(tx > MAP_WIDTH - 1) return ITEM_NULL;
    if(ty < 0) return ITEM_NULL;
    else if(ty > MAP_HEIGHT - 1) return ITEM_NULL;

    // Get the item
    if(Mode == GAMEMODE_SERVER) SDL_LockMutex(mutMap);
    tmp = tMapGame.tMapBomb[tx][ty].item;
    if(Mode == GAMEMODE_SERVER) SDL_UnlockMutex(mutMap);

    // Return
    return tmp;
}

// Set .item on the map given tx,ty and the value
void cGameField::SetItem(int tx, int ty, Uint8 item)
{
    // Check boundaries
    if(tx < 0) return;
    else if(tx > MAP_WIDTH - 1) return;
    if(ty < 0) return;
    else if(ty > MAP_HEIGHT - 1) return;

    if(Mode == GAMEMODE_SERVER) SDL_LockMutex(mutMap);
    tMapGame.tMapBomb[tx][ty].item = item;
    if(Mode == GAMEMODE_SERVER) SDL_UnlockMutex(mutMap);
}

// Initialize the network game (network or local)
// mapname with bMode=CLIENT, is the server IP
// Set the game type: random, league, league last match or timed(plus maximum time in seconds to play this match)
// Return to play with black screen
Uint8 cGameField::Init(int bMode,char *mapname,sPlayers *getPlayers,int type,int value)
{    
    // Pointer to the global players table
    tPlayers = getPlayers;

    // Set the match mode
    Mode = bMode;

    // Game mode selection
    switch(bMode)
    {
        // --------------Local mode----------------------
    case GAMEMODE_LOCAL:

        // Set the game type
        Type = type;
        if(Type == GAMETYPE_TIMED)
        {
            if(value < 0) iTimeMatch = 1;
            else iTimeMatch = value;
        }

        // Get number of players from the global var
        iNumPlayers = iPlayersLocal;		

        bServer = 0;
        strcpy(name,"TTNLocal");
        iTimeStartDelay = 0;

        // Load the map
        MapLoad(mapname);		
        ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Running on local mode with %d players.\n",iNumPlayers);

        // Launch match
        return bMapReady;
        break;

        // ---------------Client mode--------------------
    case GAMEMODE_CLIENT:
        // mapname for GAMEMODE_CLIENT is used for the server IP
        return InitClient(mapname);
        break;

        // -----------------Server mode------------------
    case GAMEMODE_SERVER:
        // Set the game type
        Type = type;
        if(Type == GAMETYPE_TIMED)
        {			 
            if(value < 0) iTimeMatch = 1;
            else iTimeMatch = value;
        }

        return InitServer(mapname);
        break;
    default:
        return GAME_END;
    }	
    return GAME_END;
}

// Initialize the match server
int cGameField::InitServer(char *mapname)
{
    Uint8 done;
    char *buff = NULL; 
    Uint32 result; 
    Uint32 size = 0, allready = 0, j = 0, i = 0, k = 0;
    sClientInfo *cInfo;

    // 0.Interfaz network initialize
    if(!INetwork->Init(LOG_FILE)) 
    {
        ILogSystem.Msg(LOG_NORMAL,"\n [GameField Error] - Couldnt initialize INetworkSystem\n");
        return GAME_ABORT; 
    }  
    if((INetwork->Features(-1) && INETFEATURE_DISABLE_LOGIN) != 1) INetwork->Features(INETFEATURE_DISABLE_LOGIN); // Disable logins

    // 1.Get number of total and remote players from the global vars
    iNumRemotePlayers = iPlayersRemote;
    iNumPlayers = iPlayersLocal + iPlayersRemote;		
    bServer = 1;
    strcpy(name,"TTNServer");	

    // 2.Load the map
    MapLoad(mapname);
    ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Running as server.\n");

    // 3.Create the server thread and our core server function
    INetwork->SetCoreServerCallback(myCoreServer);
    if(!INetwork->CreateServer(GAMESERVER_PORT,5000,0)) 
    { 
        ILogSystem.Msg(LOG_NORMAL," [GameField Error] - Couldnt initialize server.\n");
        return GAME_ABORT; 
    }
    ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Server successfully created.\n");

    // 4.Waiting for remote clients loop
    CRM32Pro.ConSetVisible(1);
    CRM32Pro.ConPrint("Starting a new network game (%s)",GetGameTypeName());
    result = INetwork->GetMyIP(); 
    CRM32Pro.ConPrint("Server listening at '%d.%d.%d.%d' port '%d'",result >> 24,(result >> 16) & 0xff,(result >> 8) & 0xff,result & 0xff,GAMESERVER_PORT);
    result = 0;
    CRM32Pro.ConPrint("Press 'Escape' key to abort");
    CRM32Pro.ConPrint("");
    CRM32Pro.ConPrint("Waiting for %d remote %s...",iNumRemotePlayers,(iNumRemotePlayers>1)?"players":"player");
    ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Waiting for the connection of %d %s",iNumRemotePlayers,(iNumRemotePlayers>1)?"players":"player");
    // 5.Connect as a client
    if((INetwork->Features(-1) && INETFEATURE_DISABLE_LOGIN) == 1) INetwork->Features(INETFEATURE_DISABLE_LOGIN); // Enable logins
    if(INetwork->ConnectTo(host,GAMESERVER_PORT,name,5000) != 1)
    {
        ILogSystem.Msg(LOG_NORMAL," [GameField Error] - Couldnt initialize client.\n");
        return GAME_ABORT;
    }
    ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Connected to local server.\n");
    done = 0;
    while(done == 0)
    {	
        // 5.1.Update internal status
        while(CRM32Pro.Update()); 

        // 5.2.Update the console background if needed (black color)
        // Todo: intentar arreglar el server entre reconexiones...
        //if(CRM32Pro.IsRenderNeeded()) SDL_FillRect(CRM32Pro.screen,CRM32Pro.ConGetRect(),0);
        SDL_FillRect(CRM32Pro.screen,NULL,0);
        CRM32Pro.RenderNeeded();

        // 5.3.Have we received data?
        result = INetwork->ReceiveData(&buff,&size);		
        switch(result)
        {	
            // Nothing to do
        case 0:
            break;
            // A new client!
        case INETWORK_NEWCLIENT: 								
            break;
            // One client left the match...
        case INETWORK_QUITCLIENT: 				
            // Check if the player was "fully" connected
            // Rejected players also reach this point but they were not fully connected.
            // We know that the name is unique for each remote player (CRM32Pro INetwork assures it) but we must check them against local players
            for(i = 0; i < 4; i++)
            {
                // Check for name matching on remote players. Local player can not disconnect this way
                if((!strcmp(tPlayers[i].name,buff)) && (tPlayers[i].remote == 1))
                {
                    // Clean this slot to allow the same player to connect
                    strcpy(tPlayers[i].name,"");
                    allready--;
                    ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Player '%s' disconnected (%d).\n",buff,allready);	
                    CRM32Pro.ConPrint("Player '%s' disconnected",buff);
                }
            }
            break;		
            // User DN (for receiving and sending the player's name and "titan")
        case INETWORK_DATA:				
            memcpy(&TBDN,buff,size);
            // IMPORTANT: free received data
            INetwork->FreeData(buff);

            // Look for TBDN_NEWPLAYER DN
            if(TBDN.type == TBDN_NEWPLAYER)
            {
                // Search in our internal tPlayers struct (the server has the right table) to check if the incoming player is remote
                // The remote client has set the ID!=0 to identify itself. Avoid also to overwrite an existing player (clients are unique but we must check against the local players)
                j = 0;
                for(i = 0; i < 4; i++)
                {						
                    if((tPlayers[i].remote == 1) && (TBDN.player[i].id != 0) && (j == 0)) 
                    {
                        // Check for names and avoid duplicated ones
                        for(k = 0; k < 4; k++)
                        {
                            // If it is a league or timed game type, after first match we avoid
                            // to reject same name players as tPlayers[] is not cleaned just checking
                            // that i != k
                            if(i != k)
                            {
                                if(!strcmp(tPlayers[k].name,TBDN.player[i].name)) j++; 
                            }
                        }

                        // With j != 0, duplicated name so abort the connection attempt..
                        if(j == 0)
                        {
                            // It is a valid remote slot, so we accept it
                            strcpy(tPlayers[i].name,TBDN.player[i].name);

                            // Send the new player struct to all clients
                            TBDN.type = TBDN_SENDPLAYERS;
                            memset(&TBDN.player,0,sizeof(sPlayers)*4);
                            memcpy(&TBDN.player,tPlayers,sizeof(sPlayers)*4);
                            result = INetwork->SendData((char *)&TBDN,sizeof(TBDN));
                            //ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Send players result: %d.\n",result);								
                            ILogSystem.Msg(LOG_NORMAL," [GameField Info] - New player '%s' connected.\n",tPlayers[i].name);	
                            CRM32Pro.ConPrint("New player '%s' connected",tPlayers[i].name);

                            // Send game type
                            TBDN.type = TBDN_GAMETYPE;
                            TBDN.data[32] = Type;
                            memcpy(&TBDN.data,&iTimeMatch,sizeof(iTimeMatch));
                            result = INetwork->SendData((char *)&TBDN,sizeof(TBDN));
                            ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Send gametype result: %d.\n",result);

                            // Send Map
                            TBDN.type = TBDN_SENDMAP;
                            memcpy(&TBDN.data,&tMapGame,sizeof(tMapGame));
                            result = INetwork->SendData((char *)&TBDN,sizeof(TBDN));
                            ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Send map result: %d.\n",result);

                            // All the data was sent!
                            j = 1;	
                        }
                        else j = 0;
                    }
                }
                ILogSystem.Msg(LOG_NORMAL," [GameField Info] - tPlayers[]:\n");
                for(i = 0; i < 4; i++)
                {
                    ILogSystem.Msg(LOG_NORMAL,"      [%d] Name: %s - ID: %d - Remote: %d\n", i, tPlayers[i].name, tPlayers[i].id, tPlayers[i].remote);
                }
                ILogSystem.Msg(LOG_NORMAL,"\n");
                // Check if the player was rejected
                if(j == 0)
                {
                    // Send a message to all clients with the name of the one who wasnt accepted. He will close his session
                    TBDN.type = TBDN_KILLPLAYER;
                    for(i = 0; i < 4; i++)
                    {
                        // The remote player trying to get access is the one in TBDN.player with id!=0
                        if(TBDN.player[i].id != 0)
                        {
                            TBDN.player[0].id = TBDN.player[i].id;
                            strcpy(TBDN.player[0].name,TBDN.player[i].name);
                            result = INetwork->SendData((char *)&TBDN,sizeof(TBDN));
                            ILogSystem.Msg(LOG_NORMAL,"Player rejected!\n");
                        }
                    }
                }
            }

            // Look for TBDN_READYPLAYER DN
            // Set that the client have all the information and it is ready for starting!
            else if(TBDN.type == TBDN_READYPLAYER)
            {
                // Inc remote player counter
                allready++;
            }

            // We should not reach this point...
            else ILogSystem.Msg(LOG_NORMAL," [GameField Error(2)] - Unknown data type received. (%d)\n",TBDN.type);
            break;
        }  
        // 5.4.IMPORTANT: free de data
        INetwork->FreeData(buff);

        // 5.6.Abort server creation
        if(CRM32Pro.GetKeystate(SDLK_ESCAPE)) 
        {
            INetwork->QueryKillServer();
            ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Aborting the match creation.\n");
            CRM32Pro.ConPrint("Aborting the match!");
            done = GAME_ABORT;				 
        }

        // 5.7.Check if all remote clients are connected.. 
        if(allready == iNumRemotePlayers)
        {
            // Get the delay for each remote player
            size = INetwork->GetClientsInfo(&cInfo);
            for(i = 0; i < 4; i++)
            {
                if((tPlayers[i].id != 0) && (tPlayers[i].remote == 1))
                {
                    for(j = 0; j < size; j++)
                    {
                        if(!strcmp(tPlayers[i].name,cInfo[j].szName))
                        {
                            tPlayers[i].iDelay = cInfo[j].iLatency;
                            break;
                        }
                    }
                }
            }
            // Calculate the delay. Set our local delay = maximum delay
            iTimeStartDelay = 0;
            for(i = 0; i < 4; i++)
            {
                if((tPlayers[i].id != 0) && (tPlayers[i].remote == 1) && (tPlayers[i].iDelay > iTimeStartDelay))
                    iTimeStartDelay = tPlayers[i].iDelay;
            }
            ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Maximum client delay of %dms\n",iTimeStartDelay);
            // Apply the delay to remote clients
            for(i = 0; i < 4; i++)
            {
                if(tPlayers[i].id != 0) tPlayers[i].iDelay = iTimeStartDelay - tPlayers[i].iDelay;
            }
            // Send start signal
            TBDN.type = TBDN_STARTMAP;
            memset(&TBDN.player,0,sizeof(sPlayers)*4);
            memcpy(&TBDN.player,tPlayers,sizeof(sPlayers)*4);
            result = INetwork->SendData((char *)&TBDN,sizeof(TBDN));
            ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Send start signal result: %d.\n\n",result);				
            done = GAME_GO;
            //s=SDL_GetTicks();
        }
    }

    // 6.Are we starting the match?
    if(done == GAME_GO)
    {
        // Inmediately close the console
        CRM32Pro.ConFeatures(CONFEATURE_DISABLE_FADE); // enable this flag
        CRM32Pro.ConSetVisible(0);
        CRM32Pro.ConFeatures(CONFEATURE_DISABLE_FADE); // disable this flag
        CRM32Pro.ConCmd("con clear");

        // Disable logins
        if((INetwork->Features(-1) && INETFEATURE_DISABLE_LOGIN) != 1) INetwork->Features(INETFEATURE_DISABLE_LOGIN); 

        // Set bServer to 2 to allow the Authoritative function starts to work
        mutMap = SDL_CreateMutex();
        bServer = 2;
    }
    return done;
}

// Initialize the client server
int cGameField::InitClient(char *server_ip)
{
    Uint8 done = 0, bMap = 0, bPlayers = 0;
    char *buff = NULL; 
    Uint32 result = 0; 
    Uint32 size = 0, allready = 0, j = 0, i = 0;
    int iRetries, iLogin;

    // 0.Interfaz network initialize
    if(!INetwork->Init(LOG_FILE)) 
    {
        ILogSystem.Msg(LOG_NORMAL,"\n [GameField Error] - Couldnt initialize INetworkSystem\n");
        return GAME_ABORT; 
    }  

    // 1.Get my name and start up the console
    // As we only allow one "remote" player per "client", here we have just to look for a name & remote = 0
    // Anyway, as a debug measure we check if there are more than one remote = 0...and abort the game
    for(i = 0; i < 4; i++) 
    {
        if((strlen(tPlayers[i].name) > 1) && (tPlayers[i].remote == 0)) 
        { 
            strcpy(name,tPlayers[i].name); 
            done++;
        }
    }
    if(done > 1) 
    {
        ILogSystem.Msg(LOG_NORMAL,"\n [GameField Error] - More than one local player detected!\n");
        return GAME_ABORT; 
    }
    ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Running as client.\n");
    bServer = 0;
    CRM32Pro.ConFeatures(CONFEATURE_DISABLE_FADE); // enable this flag
    CRM32Pro.ConSetVisible(1);
    CRM32Pro.ConFeatures(CONFEATURE_DISABLE_FADE); // disable this flag
    CRM32Pro.ConPrint("Press 'Escape' key to abort");
    while(CRM32Pro.Update()); // Render the console message

    // 2.Connect to the server and send our name and "titan"
    // Detect return error code 14 (server disabled new logins) in order to retry the connection
    iRetries = 0;    
    do
    {
        ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Player '%s' connecting to server(%d)...",name,iRetries+1);
        CRM32Pro.ConPrint("Player '%s' connecting to server(%d)...",name,iRetries+1);
        i = INetwork->ConnectTo(server_ip,GAMESERVER_PORT,name,5000);
        iRetries++;
        if(i != 1) SDL_Delay(1000); // between each retry, wait 1 second
    } while((i == 14) && (i != 1) && (iRetries < 3)); // 3 retries
    if(i != 1)
    {
        ILogSystem.Msg(LOG_NORMAL," [GameField Error] - Couldn't connect to the server(%d).\n",j+1);
        CRM32Pro.ConPrint("Could not connect to the server, aborting!");
        return GAME_ABORT;
    }

    ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Connected to '%s'.\n",host);
    CRM32Pro.ConPrint("Connected to '%s'",host);

    // 3.Print log information
    TBDN.type = TBDN_NULL;
    ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Client status: waiting for receiving the map...\n");
    ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Client status: waiting for players table...\n");
    ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Client status: waiting for start signal...\n");

    // 4.Loop to receive map/players/start signal data
    done = 0; iRetries = iLogin = 0;
    while(done == 0)
    {
        // 4.1.Update internal status
        while(CRM32Pro.Update()); 

        // 4.2.Update the console background if needed (black color)
        //if(CRM32Pro.IsRenderNeeded()) SDL_FillRect(CRM32Pro.screen,CRM32Pro.ConGetRect(),0);
        SDL_FillRect(CRM32Pro.screen,CRM32Pro.ConGetRect(),0);
        CRM32Pro.RenderNeeded();

        // 4.3.Have we received data?
        result = INetwork->ReceiveData(&buff,&size);		

        // 4.4.System datanet
        if(result != INETWORK_DATA)
        {	
            // Check system packets...
            switch(result)
            {
            case INETWORK_CLOSE:
            case INETWORK_ERROR:
                if(iRetries > 3) // Allow some attempts
                {
                    ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Server is closing. Closing this client...\n");
                    CRM32Pro.ConPrint("Server is closing. Aborting the match!");
                    done = GAME_ABORT;
                }
                else 
                { 
                    ILogSystem.Msg(LOG_NORMAL," [GameField Warning] - Received '%d' but retrying!\n",result); 
                    iRetries++; 
                }
                break;
            case INETWORK_CLIENTSINFO:
                ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Clients information updated\n");
                break;
            case INETWORK_DATAVALIDATED:
                ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Query accepted\n");
                break;
            case INETWORK_DATADENIED:
                ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Query denied\n");
                break;
                // If we are faster than the server or if we go after the server, wait for a ping and send our request!
                // This way, we assure we always go after the server creation
            case INETWORK_PING:
                if(iLogin == 0)
                {
                    iLogin = 1;
                    TBDN.type = TBDN_NEWPLAYER;
                    memset(&TBDN.player,0,sizeof(sPlayers)*4);
                    memcpy(&TBDN.player,tPlayers,sizeof(sPlayers)*4);
                    result = INetwork->SendData((char *)&TBDN,sizeof(TBDN));
                    ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Server is ready(detected a ping), sending our request(%d)\n",result);
                }
                break;				
            }

            // IMPORTANT: free received data
            INetwork->FreeData(buff);
        }	

        // 4.5.User datanet
        else
        {   
            // Receive an user datanet
            memcpy(&TBDN,buff,size);

            // IMPORTANT: free received data
            INetwork->FreeData(buff);

            // Game type, map data, player table, wait for go or error
            switch(TBDN.type)
            {
                // Gametype DN
            case TBDN_GAMETYPE:
                // Get only the first reception. Others can be sent to others clients
                if(Type == 255)
                {
                    Type = TBDN.data[32];
                    memcpy(&iTimeMatch,TBDN.data,sizeof(iTimeMatch));
                    CRM32Pro.ConPrint("Gametype set to '%s'",GetGameTypeName());
                    ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Gametype successfully received! (%d - %d)\n",Type,iTimeMatch);
                }
                break;

                // Map DN
            case TBDN_SENDMAP:
                // Get only the first reception. Others can be sent to others clients
                if(!bMap)
                {
                    ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Map successfully received!\n");	      
                    memcpy(&tMapGame,TBDN.data,sizeof(tMapGame));
                    bMap = 1;
                }
                break;

                // Players struct DN
            case TBDN_SENDPLAYERS:
                memset(tPlayers,0,sizeof(sPlayers)*4);
                memcpy(tPlayers,TBDN.player,sizeof(sPlayers)*4);
                for(i = 0; i < 4; i++)
                {
                    // Set all players to remote but we as local
                    if(!strcmp(name,tPlayers[i].name)) tPlayers[i].remote = 0;
                    else tPlayers[i].remote = 1;						
                }
                ILogSystem.Msg(LOG_NORMAL," [GameField Info] - tPlayers[]:\n");
                for(i = 0; i < 4; i++)
                {
                    ILogSystem.Msg(LOG_NORMAL,"      [%d] Name: %s - ID: %d - Remote: %d\n", i, tPlayers[i].name, tPlayers[i].id, tPlayers[i].remote);
                }
                ILogSystem.Msg(LOG_NORMAL,"\n");

                if(!bPlayers)
                {
                    ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Players table successfully received!\n");						
                }
                else
                {
                    ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Players table successfully updated!\n");
                }
                bPlayers = 1;
                break;

                // Startup signal DN
            case TBDN_STARTMAP:
                //s=ITimeSystem->GetTime();
                // Look for the local player(we) and get its delay
                // Get total number of players
                iTimeStartDelay = 0;
                for(i = 0; i < 4; i++)
                {
                    if(tPlayers[i].remote == 0) iTimeStartDelay = TBDN.player[i].iDelay;
                    if(tPlayers[i].id != 0) iNumPlayers++;
                }	
                ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Received start signal with %dms of delay.\n\n",iTimeStartDelay);	      

                // Start inmediately!
                bMapReady = 1;
                done = GAME_GO;
                break;

                // Server rejected a player 
            case TBDN_KILLPLAYER:
                // Check if we are the player rejected
                for(i = 0; i < 4; i++)
                {
                    if(!strcmp(tPlayers[i].name,name))
                    {
                        j = tPlayers[i].id; break;
                    }
                }
                if((!strcmp(TBDN.player[0].name,name)) && (TBDN.player[0].id == j))
                {
                    ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Server rejected this player.\n");
                    CRM32Pro.ConPrint("Server rejected the connection");
                    INetwork->QueryKillClient();
                }
                break;

                // With additional players, we will receive this DN but nothing to do with it
            case TBDN_NEWPLAYER:
                break;
                // Unknown DN
            default:
                ILogSystem.Msg(LOG_NORMAL," [GameField Error] - Unknown data type received(%d).\n",TBDN.type);
                break;
            }
        }

        // 4.6.Abort client connection
        if(CRM32Pro.GetKeystate(SDLK_ESCAPE)) 
        {
            INetwork->QueryKillClient();
            ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Aborting the game creation.\n");
            CRM32Pro.ConPrint("Aborting the match!");
            done = GAME_ABORT;				 
        }

        // 4.7.Have we received the map, the game type and the players table? -> send to the server we are ready!
        if((bMap == 1) && (Type != 255) && (bPlayers == 1))
        {
            TBDN.type = TBDN_READYPLAYER;
            memset(&TBDN.player,0,sizeof(sPlayers)*4);
            memcpy(&TBDN.player,tPlayers,sizeof(sPlayers)*4);
            result = INetwork->SendData((char *)&TBDN,sizeof(TBDN));
            ILogSystem.Msg(LOG_NORMAL," [GameField Info] - We are ready!, waiting for the server startup signal\n");
            CRM32Pro.ConPrint("Waiting for the server startup signal...");
            bMap = 2; // Avoid to reenter
        }
    }  

    // Are we starting the match?
    if(done == GAME_GO)
    {
        // Inmediately close the console
        CRM32Pro.ConFeatures(CONFEATURE_DISABLE_FADE); // enable this flag
        CRM32Pro.ConSetVisible(0);
        CRM32Pro.ConFeatures(CONFEATURE_DISABLE_FADE); // disable this flag
        CRM32Pro.ConCmd("con clear");
    }
    return done;
}

// Authorization callback server. As this will run on a separated thread, all shared date must be secured using a mutex.
// We process bombs creation(with possible movement), attempts for catching an item and death queries
int  cGameField::myCoreServer(void *data,int size)
{
    sTBDN_Event *myEvent;
    int iXX = -1, iYY = -1;
    int iOutOfSync = 0;

    // 1.Checks: a valid pointer to the game field, are we on the right bServer status? is a valid DN?
    if(myGameField == NULL) return 0;
    if(myGameField->bServer != 2) return 1;
    if(size != sizeof(sTBDN_Event)) return 1;

    // 2.Pointer to the real data in order to be able to modify it
    myEvent = (sTBDN_Event *)data;

    // 3.Check sync between client<->server and get the server side coordinates and set out of sync var
    for(int i = 0; i < 4; i++)
    {
        // Look for the player event on our local tPlayer[]
        if(myGameField->tPlayers[i].id == myEvent->iCtrlID)
        {
            iXX = myGameField->tPlayers[i].player->GetXr();
            iYY = myGameField->tPlayers[i].player->GetYr();
            if((myEvent->event.Xr != iXX) || (myEvent->event.Yr != iYY))			
            { 					
#ifdef DEBUG_CORESERVER			
                ILogSystem.Msg(LOG_NORMAL," [CoreServer Error] - Position '%d' mismatch: Server (%d,%d) vs Client (%d,%d)\n",
                    myEvent->iCtrlID, iXX, iYY, myEvent->event.Xr,myEvent->event.Yr);
#endif

                // set out of sync var on
                iOutOfSync = 1;
            }
            // break the search, there is only one event per player :)
            break;
        }
    }		

    // 4.Check bomb creation at server tPlayer[] current position. The bomb creation requires a synchronized event 
    // but if the event includes a movement..we accept it removing the bomb creation :)
    if(myEvent->event.bomb == 1) 
    {	
        // Out of sync?
        if(iOutOfSync)
        {
            // Check if the event includes a movement, if so, remove the bomb creation and accept the movement
            if((myEvent->event.move_x != 0) || (myEvent->event.move_y != 0))
            {
                myEvent->event.bomb = 0; 

#ifdef DEBUG_CORESERVER			
                ILogSystem.Msg(LOG_NORMAL," [CoreServer Error] - Bomb denied but movement accepted! for '%d' - %d\n",myEvent->iCtrlID,myEvent->event.bomb);
#endif
            }
            // It was only a bomb creation request, deny it!
            else return 0;
        }
        // The server map says where we really are
        else if(myGameField->GetStatus(iXX, iYY) <= STATUS_NOWALKABLE)
        {
#ifdef DEBUG_CORESERVER
            ILogSystem.Msg(LOG_NORMAL," [CoreServer Info] - Bomb creation query at (%d,%d) denied to player '%d'\n",iXX,iYY, myEvent->iCtrlID);
#endif		    
            // Deny the bomb creation
            return 0;
        }
        // Accept the bomb creation
        else 
        {
            // Set the cell as busy (bombcreate will create the bomb)
            myGameField->SetStatus(iXX, iYY, STATUS_NOWALKABLE);
        }
    }

    // 5.Check item query at server tPlayer[] current position. Only accepts it when we are synchronized.
    else if((myEvent->event.item != 0) && (!iOutOfSync))
    {
        if(myEvent->event.item != myGameField->GetItem(iXX,iYY))
        {
#ifdef DEBUG_CORESERVER
            ILogSystem.Msg(LOG_NORMAL," [CoreServer Info] - ItemCatch query at (%d,%d) denied to player '%d'\n",iXX, iYY, myEvent->iCtrlID);
#endif
            // Deny this event
            myEvent->event.item = 0;
            return 0;
        }

        // Accept the query and remove the item
        myGameField->SetItem(iXX,iYY,ITEM_NULL);
    }

    // 6.Check death query at server tPlayer[] current position. Only accepts it when we are synchronized.
    else if((myEvent->event.death != 0) && (!iOutOfSync))
    {
        Uint8 myRet;
        myRet = myGameField->GetStatus(iXX,iYY);

#ifdef DEBUG_CORESERVER
        if(myEvent->event.death != myRet)
        {
            ILogSystem.Msg(LOG_NORMAL," [CoreServer Info] - Death query at (%d,%d) mismatch: '%d'/'%d'. Requested by '%d' player\n",iXX, iYY, myEvent->event.death,myRet,myEvent->iCtrlID);
        }
#endif
        // Check real burning status, if it is not burning, deny the event. Otherwise, send the right burning status
        if(myRet >= STATUS_BURN_P1)
        {
#ifdef DEBUG_CORESERVER
            ILogSystem.Msg(LOG_NORMAL," [CoreServer Info] - Death query at (%d,%d) accepted and set the burning status to '%d'. Requested by '%d' player\n",iXX, iYY, myRet,myEvent->iCtrlID);
#endif
            // Accept this event but fix the burning status
            myEvent->event.death = myRet;
        }
        // Deny the event
        else 
        {
#ifdef DEBUG_CORESERVER
            ILogSystem.Msg(LOG_NORMAL," [CoreServer Info] - Death query at (%d,%d) denied. Requested by '%d' player\n",iXX, iYY, myEvent->iCtrlID);
#endif
            return 0;
        }
    }

    // 7.Deny the event
    else 
    {
        ILogSystem.Msg(LOG_NORMAL," [CoreServer Error] - Deny event!\n");
        return 0;
    }

    // 8.Accept the event
    ILogSystem.Msg(LOG_NORMAL," [CoreServer Info] - Accept event!\n");
    return 1;
}

// Map init loading all resources
void cGameField::MapInit()
{ 
    Uint32 i=0,j=0,x=0,y=0,k=0;
    SDL_Rect rect; 

    rect.w=rect.h=rect.x=rect.y=0;

    // Is the map ready?
    if(!bMapReady) return;

    // Load resources
    sGameBg = IImage->Load(GFX_RESOURCE,tMapGame.sBg);
    sBlock = IImage->Load(GFX_RESOURCE,tMapGame.sBlock);
    tIcons = new CRM32Pro_CTile;
    tIcons->Load(GFX_RESOURCE,"tIconos");

    // Master sprites: for breakables and items
    sprBreakableMaster=new CRM32Pro_CSprite;
    sprBreakableMaster->Load(GFX_RESOURCE,tMapGame.sBreakable);
    sprBreakableMaster->AutoRestore(0);
    sprBreakableMaster->Pause();
    sprItemMaster=new CRM32Pro_CSprite;
    sprItemMaster->Load(GFX_RESOURCE,"Items");
    sprItemMaster->AutoRestore(0);
    for(i=0;i<tMapGame.iNumBK;i++) 
    {
        tBreakables[i].sprItem=new CRM32Pro_CSprite;
        tBreakables[i].sprItem->Clone(sprItemMaster);
        tBreakables[i].sprItem->AutoRestore(0);
        tBreakables[i].sprBreakable=new CRM32Pro_CSprite;
        tBreakables[i].sprBreakable->Clone(sprBreakableMaster);
        tBreakables[i].sprBreakable->AutoRestore(0);
        tBreakables[i].sprBreakable->Pause(); 
        tBreakables[i].iItem=0;
    }

    // Generate the breakables
    for(i=0;i<15;i++)
        for(j=0;j<11;j++)
        {
            // Is there any?
            if(GetItem(i,j) >= ITEM_BREAKABLE)
            {			  
                // absolute position on screen
                x=(i*DELTA_X)+(Sint16)ORIGIN_X; 
                y=(j*DELTA_Y)+(Sint16)ORIGIN_Y;
                // Look for a free slot on the breakable's table
                k=0;
                while((tBreakables[k].iItem!=0) && (k<MAX_BREAKABLES)) k++;			 
                // Check for a possible error
                if(k==MAX_BREAKABLES) { i=15; j=11; break; }
                // Enable the breakable
                tBreakables[k].iStatus=1;
                tBreakables[k].iItem = GetItem(i,j);
                tBreakables[k].iX=i;
                tBreakables[k].iY=j;
                tBreakables[k].sprBreakable->SetPosition(x,y);
                tBreakables[k].sprItem->SetPosition(x,y);
                // Select the anim
                if(tBreakables[k].iItem>ITEM_BREAKABLE) 
                    tBreakables[k].sprItem->SelectAnim(SPRSTATE_NORMAL+(tBreakables[k].iItem-3)); // Subtract 3
            }
        }
}

// Libera los recursos del mapa
void cGameField::MapFree()
{
    int i;

    // Comprueba que se haya iniciado el mapa
    if(!bMapReady) return;

    // Eliminamos background, bloque
    if(sGameBg != NULL) { SDL_FreeSurface(sGameBg); sGameBg = NULL; }
    if(sBlock != NULL) { SDL_FreeSurface(sBlock); sBlock = NULL; }
    if(tIcons != NULL) { delete tIcons; tIcons = NULL; }

    // Eliminamos los sprites breakables e items, tanto clones como maestros
    for(i=0;i<tMapGame.iNumBK;i++)
    {
        tBreakables[i].iX=250;
        tBreakables[i].iY=250;
        tBreakables[i].iItem=ITEM_NULL;
        tBreakables[i].iStatus=STATUS_WALKABLE;
        if(tBreakables[i].sprItem!=NULL) delete tBreakables[i].sprItem;
        if(tBreakables[i].sprBreakable!=NULL) delete tBreakables[i].sprBreakable;
    }
    if(sprItemMaster!=NULL) delete sprItemMaster;
    if(sprBreakableMaster!=NULL) delete sprBreakableMaster;    
}

// Ejecucion logica de los breakables
void cGameField::BreakablesExecute()
{
    int k;

    // Comprueba que se haya iniciado el mapa
    if(!bMapReady) return;

    // Recorremos lista de breakables viendo a cada uno
    for(k=0;k<tMapGame.iNumBK;k++)
    {
        // Tenemos un breakable! 
        if(tBreakables[k].iItem>=ITEM_BREAKABLE)
        {		
            switch(tBreakables[k].iStatus)
            {
                // Ha recibido una orden de explosion..
            case 4:
                // Para sincronizar, detectamos el final de la explosion cuando termine la animacion del sprite
                if(tBreakables[k].sprBreakable->IsFinished())
                {
                    // Acciones comunes
                    if(Mode == GAMEMODE_SERVER) SDL_LockMutex(mutMap);
                    tMapGame.tMapBomb[tBreakables[k].iX][tBreakables[k].iY].status = STATUS_WALKABLE;
                    if(Mode == GAMEMODE_SERVER) SDL_UnlockMutex(mutMap);
                    if(tBreakables[k].sprBreakable!=NULL) { delete tBreakables[k].sprBreakable; tBreakables[k].sprBreakable=NULL; }

                    // Si se trata de un solo bloque breakable sin item, eliminamos el breakable
                    if(tBreakables[k].iItem==ITEM_BREAKABLE)
                    {
                        SetItem(tBreakables[k].iX,tBreakables[k].iY,ITEM_NULL);
                        tBreakables[k].iX=251;	tBreakables[k].iY=251; 
                        tBreakables[k].iItem=ITEM_NULL; tBreakables[k].iStatus=STATUS_WALKABLE;
                        if(tBreakables[k].sprItem!=NULL) { delete tBreakables[k].sprItem; tBreakables[k].sprItem=NULL; }
                    }
                    // Tiene un item
                    else
                    {
                        // Pasamos al siguiente estado..el que controla un item
                        tBreakables[k].iStatus=5;
                    }
                }
                break;
                // Un item que es eliminado, ya sea por bomba o por algun objeto que lo ha "atrapado"
            case 6:					 
                SetItem(tBreakables[k].iX,tBreakables[k].iY,ITEM_NULL);
                tBreakables[k].iX=251;	tBreakables[k].iY=251; 
                tBreakables[k].iItem=ITEM_NULL; tBreakables[k].iStatus=STATUS_WALKABLE;
                if(tBreakables[k].sprItem!=NULL) { delete tBreakables[k].sprItem; tBreakables[k].sprItem=NULL; }
                break;

            }
        }		 
    }
}

// Mata por explosion a un breakable,devuelve el estado asignado
Uint8 cGameField::BreakablesKill(int x,int y)
{
    int k;

    // Comprueba que se haya iniciado el mapa
    if(!bMapReady) return 0;

    // Recorremos lista de breakables buscando el objetivo.
    for(k=0;k<tMapGame.iNumBK;k++)
    {
        // Tenemos un breakable o un item 
        if(tBreakables[k].iItem >= ITEM_BREAKABLE)
        {			
            if(tBreakables[k].iX == x)
            {
                if(tBreakables[k].iY == y) 
                {
                    // Comprueba si el breakable ha sido bombardeado justo antes..
                    if(tBreakables[k].iStatus == 1) { tBreakables[k].iStatus = 4; return 4; } // Ordena un kill del breakable
                    // Comprueba si el item ha sido bombardeado justo antes..
                    if(tBreakables[k].iStatus == 5) { tBreakables[k].iStatus = 6; return 6; } // Ordena un kill del item
                }
            }
        }
    }
    // Nada
    return 0;
}

// Mata de un golpe, a todos los breakables
void cGameField::BreakablesKillAll()
{
    int j,k;

    // Comprueba que se haya iniciado el mapa
    if(!bMapReady) return;

    for(k=0;k<15;k++)
        for(j=0;j<11;j++) BreakablesKill(k,j);
}

// Initialize the players control
// Remote clients always are managed with cursors
void cGameField::PlayersInit()
{
    int i;
    int iKeySet = 0;

    if(tPlayers == NULL) 
    {
        ILogSystem.Msg(LOG_NORMAL," [GameField Error] - Can not initialize the players!");
        return;
    }

    for(i = 0; i < 4; i++)
    {
        if(strcmp(tPlayers[i].name,""))
        {
            switch(tPlayers[i].id)
            {
            case PLAYER_WATER:
                tPlayers[i].player = new cPlayer("water",&tPlayers[i],this);
                break;
            case PLAYER_FIRE:
                tPlayers[i].player = new cPlayer("fire",&tPlayers[i],this);
                break;
            case PLAYER_EARTH:
                tPlayers[i].player = new cPlayer("earth",&tPlayers[i],this);
                break;
            case PLAYER_AIR:
                tPlayers[i].player = new cPlayer("air",&tPlayers[i],this);
                break;
            }			
            if(tPlayers[i].remote == 0) { tPlayers[i].player->SetCtrlKeys(iKeySet); iKeySet++; }
        }
    }
}

// Free player objects
void cGameField::PlayersFree()
{
    int i;
    for(i=0;i<4;i++)
    {
        if(tPlayers[i].player != NULL) { delete tPlayers[i].player; tPlayers[i].player = NULL; }
    }
}

// Get event for a player from the queue (if any)
sTBDN_Event *cGameField::GetEventForPlayer(int numplayer)
{
    sTBDN_Event *my_event = NULL;

    if(!TBDN_Events[numplayer].empty())
    {
        my_event = &TBDN_Events[numplayer].front();
        TBDN_Events[numplayer].pop();
    }
    return my_event;
}

// Render a special text effect
void cGameField::SpecialText(SDL_Surface *dst_surf,CRM32Pro_CFont *font,int x, int y, char *sText, int count)
{
    double angle = 0;
    int i,xpos,ypos;
    char sChar[2];

    // Count modulates the effect
    angle = (double)count * 5.0;

    // One iteration per each character on the string
    for (i=0; i < (int)strlen(sText); i++)
    {
        sChar[0] = sText[i];
        sChar[1] = '\0';
        if (i==0)	xpos = x;
        ypos = (int)(y + sin(M_PI/180 * (angle + i*12)) * font->GetSurface()->h);
        font->PutString(dst_surf,xpos,ypos,sChar);
        xpos = xpos + font->StringWidth(sChar);
    }
}

// Show the message at the beginning of the match
// fade in. The delay is important on network matches so we try to tune it
int cGameField::MessageStartGame()
{
    Uint32 iTime = 0, iTmp, iStart, iDelay;
    Uint8 done;
    SDL_Event keyevent;

    iSpecialTextCounter = 0;   
    strcpy(sMsg,"Starting match in");
    done = 4; // Number of seconds 
    iDelay = (done * 1000) + iTimeStartDelay;
    iRenderGraphicsStatus = RENDERGRAPHICS_START;
    ISoundFX->SoundPlay(-1,hndS_startGame,0);		
    iStart = (Uint32)ITimeSystem->GetTime();

    while(done > 0)
    {
        // Logic update
        iTmp = (Uint32)ITimeSystem->GetTime();
        if(iTime == 0) 
        {
            iTime = iTmp;			
            _itoa(done,sTmp,10);
        }
        else if(iTmp > (iTime + 1000)) 
        { 
            iTime = 0; 
            done--; 
        }
        iSpecialTextCounter += 2;
        // Try to avoid too much error propagation on the delay (it will never be equal to iDelayStep)
        // As soon as the total time spent here is greater than the delay, we return back
        if(iTmp > (iStart + iDelay)) done = 0;

        // CRM32Pro update, events, render callback, etc.
        while(CRM32Pro.Update(&keyevent) && (done > 0))
        {
            switch(keyevent.type)
            {
            case SDL_KEYUP:
                switch(keyevent.key.keysym.sym)
                {
                case SDLK_ESCAPE:
                    // Only local modes or the server can reach this point as of now (clients can not abort the game) 
                    if((Mode == GAMEMODE_SERVER) || (Mode == GAMEMODE_LOCAL)) return GAME_ABORT;
                    break;
                }
                break;
            }
        } 		
    }

    return GAME_GO;
}

// Show a message at the end of the match
// It receives as parameter: GAME_GO, GAME_END or GAME_ABORT
void cGameField::MessageEndGame(int bFlag)
{
    SDL_Event keyevent;
    int iMaxPoints, iTie, done, i, j, iTime, bBubbleSort;

    // 1.Create the total score table and, if applies get the match/game winner
    strcpy(szWinner,""); // By default, no winner
    iMaxPoints = iTie = 0;
    for(j = 0; j < 4; j++)
    {
        if(tPlayers[j].player != NULL)
        {
            // Copy to our score table
            strcpy(stScore[j].sName,tPlayers[j].name);
            stScore[j].iPoints = tPlayers[j].points;
            // With GAME_GO, we get the match winner
            if(bFlag == GAME_GO)
            {
                if(tPlayers[j].player->GetPoints() > iMaxPoints) { iMaxPoints = tPlayers[j].player->GetPoints(); strcpy(szWinner, tPlayers[j].name); }
            }
            // With GAME_END, we get the game winner
            else if(bFlag == GAME_END)
            {
                if(stScore[j].iPoints > iMaxPoints) { iMaxPoints = stScore[j].iPoints; strcpy(szWinner, stScore[j].sName); }
            }
        }
        else { strcpy(stScore[j].sName,""); stScore[j].iPoints = 0; }
    }	

    // 2.Check for a tie (only if sWinner != "")
    if(strlen(szWinner) > 0)
    {
        for(j = 0; j < 4; j++)
        {
            // With GAME_GO, we look for a possible tie match
            if(bFlag == GAME_GO)
            {
                if(tPlayers[j].player != NULL)
                {
                    if(tPlayers[j].player->GetPoints() == iMaxPoints) iTie++;
                }
            }
            // With GAME_END, we look for a possible tie game
            else if(bFlag == GAME_END)
            {
                if(strcmp(stScore[j].sName,""))
                {
                    if(stScore[j].iPoints == iMaxPoints) iTie++;
                }
            }
        }
    }
    if(iTie > 1) strcpy(szWinner, ""); // No winner, it is a tie

    // 3.Prepare output messages
    switch(bFlag)
    {
    case GAME_ABORT:
        strcpy(szWinner, "");
        strcpy(sMsg,"Game aborted");
        ISoundFX->SoundPlay(-1,hndS_abortGame,0);
        break;

    case GAME_GO:
        if(strlen(szWinner) > 0) 
        {
            strcpy(sMsg,"won the match!");
            ISoundFX->SoundPlay(-1,hndS_winGame,0); 
        }
        else 
        {
            strcpy(sMsg,"tie match!");
            ISoundFX->SoundPlay(-1,hndS_tieGame,0);
        }
        break;

    case GAME_END:
        if(strlen(szWinner) > 0) 
        {
            strcpy(sMsg,"won the game!");
            ISoundFX->SoundPlay(-1,hndS_winGame,0); 
        }
        else 
        {
            strcpy(sMsg,"tie game!");
            ISoundFX->SoundPlay(-1,hndS_tieGame,0);
        }
        break;	
    }

    // 4.Sort our total score table using a bubble sort algorithm
    bBubbleSort = 1;
    for(i = 1; (i <= 4) && bBubbleSort; i++)
    {
        bBubbleSort = 0;
        for(j = 0; j < (4 - 1); j++)
        {
            if(stScore[j+1].iPoints > stScore[j].iPoints)
            { 
                strcpy(sTmp,stScore[j].sName);
                iMaxPoints = stScore[j].iPoints;             
                stScore[j].iPoints = stScore[j+1].iPoints;
                strcpy(stScore[j].sName,stScore[j+1].sName);
                stScore[j+1].iPoints = iMaxPoints;
                strcpy(stScore[j+1].sName,sTmp);
                bBubbleSort = 1;               
            }
        }
    }

    // 5.End match message loop
    iRenderGraphicsStatus = RENDERGRAPHICS_END;
    iSpecialTextCounter = 0;
    iTime = ITimeSystem->GetTime();
    done = 0;		
    while(done != 2)
    {
        // Logic update
        iSpecialTextCounter += 2;

        // Check for 5sec limit on "end of matches" or "aborted" games
        if(iTime + 5000 < ITimeSystem->GetTime()) done = 1;
        if((bFlag != GAME_END) && (done == 1)) done = 2;			

        while(CRM32Pro.Update(&keyevent)) 
        {
            switch(keyevent.type)
            {
            case SDL_KEYUP:
                // With any key and end of games, we finish the game
                if((bFlag == GAME_END) && (done == 1)) done = 2;
                break;
            }
        }		
    }
}

// Main execution of the match
Uint8 cGameField::Execute()
{  
    Uint8 done,value,i,j,k, tRandom[4];
    int iMatchEndSignal, iGameFlag;
    SDL_Event keyevent;

    // Check that the map is ready
    if(!bMapReady) return GAME_ABORT;

    // Fade out music
    ISoundFX->MusicFadeOut(500);

    // Prepare the screen	
    CRM32Pro.SetRenderCallback(myRenderInGame);
    ICursor->Hide();
    IScreenFX->RenderFX(CRM32Pro.screen,CRM32Pro.screen,FXSCREEN_GREY);
    IScreenFX->FadeToColor(0,0,0,1000);	

    // Initialize the match field
    MapInit();
    PlayersInit();	
    iMatchFSM = iTimeEOM = iMatchEndSignal = iGameFlag = 0;

    // Play music for this gamefield
    switch(tMapGame.iOwner)
    {
    case PLAYER_WATER:
        ISoundFX->MusicPlay(hndM_water, 100);
        break;

    case PLAYER_FIRE:
        ISoundFX->MusicPlay(hndM_fire, 100);
        break;

    case PLAYER_EARTH:
        ISoundFX->MusicPlay(hndM_earth, 100);
        break;

    case PLAYER_AIR:
        ISoundFX->MusicPlay(hndM_air, 100);
        break;
    }

    // Starting in 3 seconds plus the delay
    done = 0;
    //f = SDL_GetTicks();
    //ILogSystem.Msg(LOG_NORMAL,"\n -> Starting match - Real wanted delay %d - Server provided delay %d - Player number %d\n", f-s,iTimeStartDelay,iNumPlayers);
    if(MessageStartGame() == GAME_ABORT) 
    {
        iMatchFSM = 3; // Only local modes or the server can reach this point as of now (clients can not abort the game)
        iGameFlag = GAME_ABORT;
    }
    //ILogSystem.Msg(LOG_NORMAL," -> Starting match - Real delay %d (diff %d)\n", SDL_GetTicks() - s, SDL_GetTicks() - s - (f-s));

    // -----Main match loop-----
    CRM32Pro.CleanUp();
    iTimeStart = ITimeSystem->GetTime();
    iRenderGraphicsStatus = RENDERGRAPHICS_GAME;
    while(done == 0)
    {
        // 1.With network matches, received all DNs storing the events on queues (FIFO lists) for each player
        if(Mode != GAMEMODE_LOCAL)
        {
            char *buff=NULL; 
            Uint32 size=0;
            Uint32 status;
            status = 1;
            while(status != 0)
            {
                TBDN_Event.iCtrlID = TBDN_NULL;	    
                // Check DN with Events from other players or my own data validated
                status = INetwork->ReceiveData(&buff,&size);
                if(((status == INETWORK_DATA) || (status == INETWORK_DATAVALIDATED)) && (size == sizeof(sTBDN_Event)))
                {
                    // Copy data to our struct
                    memcpy(&TBDN_Event,buff,size);
                    // Process the event
                    // Look for player events and as soon as we find one, push it on the player queue a break the loop
                    // Note that we also receive our own events. They were sent to be validated
                    // We also process match logic control event (TBDN_GAMENEXT and TBDN_GAMEEND)
                    for(i = 0; i < 4; i++)
                    {
                        if(tPlayers[i].id == TBDN_Event.iCtrlID) 
                        {
                            TBDN_Events[i].push(TBDN_Event);
                            break;
                            // With debug mode, it dumps all the events for each player for easy debugging
                            /*if(bDebug)
                            {
                            FILE *pF;
                            // Creation of the files the first time
                            if(fp1 == NULL) fp1 = fopen("player1.txt","wt"); if(fp2 == NULL) fp2 = fopen("player2.txt","wt");
                            if(fp3 == NULL) fp3 = fopen("player3.txt","wt"); if(fp4 == NULL) fp4 = fopen("player4.txt","wt");
                            // Dumps info
                            switch(i)
                            {
                            case 0: pF = fp1; break; case 1: pF = fp2; break; case 2: pF = fp3; break; case 3: pF = fp4; break;									
                            }
                            fprintf(pF,"move_x %d - move_y %d - Xr %d - Yr %d - bomb %d - item %d - points %d - death %d\n",TBDN_Event.event.move_x,TBDN_Event.event.move_y, TBDN_Event.event.Xr, TBDN_Event.event.Yr,TBDN_Event.event.bomb, TBDN_Event.event.item, TBDN_Event.event.points, TBDN_Event.event.death);

                            }*/
                        }
                    }
                    // It isnt a player event so it must be a match logic control
                    if(i == 4)
                    {
                        ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Game logic control event (%d,%d)\n",TBDN_Event.iCtrlID,TBDN_Event.event.move_x);
                        iGameFlag = iMatchEndSignal = TBDN_Event.iCtrlID;
                        iMatchFSM = TBDN_Event.event.move_x;
                    }
                }
                // With the Ping signal, we request an update of the latency
#ifdef DEBUG_NETWORK
                else if(status == INETWORK_PING)
                {
                    // Only request it, local client (to avoid each remote client to query it)
                    if(Mode == GAMEMODE_SERVER) INetwork->QueryClientsInfo();
                }
#endif
                // With these errors, abort this game
                else if((status == INETWORK_ERROR) || (status == INETWORK_CLOSE) || (status == INETWORK_QUITCLIENT))
                {
                    ILogSystem.Msg(LOG_NORMAL," [GameField Error] - Detected network problems, aborting the game! (%d)\n",status);
                    iGameFlag = iMatchEndSignal = GAME_ABORT;
                    iMatchFSM = 4;
                    status = 0; // To exit from the reception loop
                }
                // Other status as INETWORK_DATADENIED are discarded so free the memory used:
                INetwork->FreeData(buff);
            }
        }

        // 2.Items execution
        BreakablesExecute();

        // 3.Create a random player execution table
        for(k = 0; k < 4; k++) tRandom[k] = k;
        for(k = 4; k > 1; k--)
        {
            i = CRM32Pro.Rand() % k;
            j = tRandom[i];
            tRandom[i] = tRandom[k - 1];
            tRandom[k - 1] = j;
        }

        // 4.Random execution of the players with theirs bombs
        for(k = 0; k < 4; k++)
        {		
            i = tRandom[k];

            if(tPlayers[i].player != NULL) 
            {
                // Control: movement and bomb creation for local/remote players
                tPlayers[i].player->Control(i);

                // Execution of each player (bombexecute,ItemCatch and Dead logic)
                value = tPlayers[i].player->Execute();

                // Are we alive?
                if(value != PLAYER_ALIVE)
                {
                    // Check for an autokill
                    if(tPlayers[i].id & value) SetPoints(tPlayers[i].id,POINTS_AUTOKILL,tPlayers[i].player->GetXr(),tPlayers[i].player->GetYr());
                    // Someone killed me
                    else SetPoints(value,POINTS_KILLP,tPlayers[i].player->GetXr(),tPlayers[i].player->GetYr());

                    // We lost all our lifes :(
                    if(value & PLAYER_DEATH) iNumPlayers--;
                }
            }
        }	  

        // 5.If we are on a timed match... check for the final
        if(iTimeMatch > 0)
        {
            k = (ITimeSystem->GetTime() - iTimeStart) / 1000;
            if(iTimeMatch - k < 30)
            {
                k = iTimeMatch - k;
                _itoa(k,sCounterDown,10);
                // Detect end of time
                if(k == 0) 
                {
                    // Only on the server or on local mode, it start the stopping procedure. The clients will receive the signal
                    // Network match, send the signal
                    if((Mode == GAMEMODE_SERVER) || (Mode == GAMEMODE_LOCAL))
                    {	
                        iMatchFSM = 2;
                        iGameFlag = GAME_END;
                    }					
                    // Disable the counterdown. Set we are running out of time
                    iTimeMatch = -1;
                }
            }
        }

        // 6.Check for the end of match. 
        // Let continue for TIME_EOM since only 1 or 0 players left
        if((iNumPlayers <= 1) || (iGameFlag != 0))
        {
            // Small End Of Match FSM
            switch(iMatchFSM)
            {
            case 0: // Local, client or server
                iTimeEOM = ITimeSystem->GetTime();
                iMatchFSM++;
                break;

            case 1: // Local, client or server. 
                // The client wait here for the server iMatchEndSignal.
                // Server or local wait TIME_EOM
                if((Mode != GAMEMODE_CLIENT) && (iTimeEOM + TIME_EOM < ITimeSystem->GetTime())) iMatchFSM++;
                break;

            case 2: // Only local or server
                // One player is alive after TIME_EOM
                if(iNumPlayers == 1)
                {
                    // Look for the player alive...
                    j = k = 0;
                    while(k < 4)
                    {
                        if(tPlayers[k].player != NULL)
                        {
                            if(tPlayers[k].player->Alive()) { j = k; k = 4; }
                        }
                        k++;
                    }
                    // Apply right bonus...
                    if(tPlayers[j].id != tMapGame.iOwner) 
                    {
                        // A visitor won the match!
                        tPlayers[j].player->AddPoints(POINTS_VISITORWIN,0,tPlayers[j].player->GetXr(),tPlayers[j].player->GetYr()); 
                        ILogSystem.Msg(LOG_NORMAL," -> Only '%s' visitor player survived. Applying special bonus\n",tPlayers[j].name);
                    }
                    else
                    {
                        // Local player won the match!
                        tPlayers[j].player->AddPoints(POINTS_LOCALWIN,0,tPlayers[j].player->GetXr(),tPlayers[j].player->GetYr());
                        ILogSystem.Msg(LOG_NORMAL," -> Only '%s' local player survived. Applying normal bonus\n",tPlayers[j].name);                                
                    }							
                    // Another bonus, an extra life for the next map
                    tPlayers[j].bonus = 1;					
                }
                // All players died so no bonus at all
                else if(iNumPlayers == 0) ILogSystem.Msg(LOG_NORMAL," -> All players died. There is not bonus for anyone!\n");  
                // End of match by time
                else 
                {
                    ILogSystem.Msg(LOG_NORMAL," -> End of match detected and more than one player survived. There is not bonus for anyone!\n");
                }

                // Go to next state
                iMatchFSM++;
                break;

            case 3:	// Only local or server
                // Create the match end signal (end? abort? is there another match?)
                // If iGameFlag = 0, there is only one or none alive player so we close this match and continue the game (GAME_GO)
                // only if this is not the last match of a league or a random match (GAME_END)
                if(iGameFlag == 0)
                {
                    if((Type == GAMETYPE_LEAGUE_LAST) || (Type == GAMETYPE_RANDOM)) iGameFlag = GAME_END;
                    else iGameFlag = GAME_GO;
                }
                // If iGameFlag = GAME_END, it is a finished timed match
                // If iGameFlag = GAME_ABORT, it is cancelled game

                // If I am the server, I propagate the match end signal to all clients connected
                // and wait for receiving this event... . Set the gamestatus to 4 to avoid infine-loop
                if(Mode == GAMEMODE_SERVER) 
                {
                    SendMatchEndSignal(iGameFlag, 4);					
                }
                // If this is a local game, set everything to finish
                else if(Mode == GAMEMODE_LOCAL) 
                {
                    iMatchEndSignal = iGameFlag;
                    ILogSystem.Msg(LOG_NORMAL," [GameField Info] - Game logic control event (%d)\n",iMatchEndSignal);
                }
                iMatchFSM++;
                break;				

            case 4:	// Local, client or server
                // Have we the match end signal?
                if(iMatchEndSignal != 0)
                {
                    // Flush the player events looking for points event as it could be possible to 
                    // receive the points with the winner and the end message at the same time
                    // Add current player points to its general points (for the high score table)
                    // If the player is still alive, set it on survival mode
                    for(j = 0; j < 4; j++)
                    {
                        if(tPlayers[j].player != NULL) 
                        {
                            int iFlushEv = 0;
                            sTBDN_Event *pTBDN_Event = NULL;
                            while(!iFlushEv)
                            {
                                iFlushEv = 1;
                                pTBDN_Event = GetEventForPlayer(j);
                                if(pTBDN_Event != NULL)
                                {
                                    // Get points validations
                                    if(pTBDN_Event->event.points != 0) 
                                    { 
                                        tPlayers[j].player->AddPoints(pTBDN_Event->event.points,1,pTBDN_Event->event.Xr,pTBDN_Event->event.Yr); 
                                        iFlushEv = 0; // retry again looking for points
                                    }
                                    // Deaths and items are discarded
                                    if((pTBDN_Event->event.item != ITEM_NULL) || (pTBDN_Event->event.death != 0)) iFlushEv = 0; // retry again looking for points
                                }
                            }
                            // Save current match points to our general table
                            tPlayers[j].points = tPlayers[j].points + tPlayers[j].player->GetPoints();	
                            if(tPlayers[j].player->Alive()) tPlayers[j].player->Survive();
                        }
                    }

                    // Final animation showing the winner of this match(the player with more points, not the one who survived).
                    fFPS = ITimeSystem->GetRFR();
                    iTimeEnd = ITimeSystem->GetTime(); // End of play time (time on end message is not accounted)
                    ISoundFX->MusicFadeOut(500); // Fade out music
                    MessageEndGame(iMatchEndSignal);
                    done = iMatchEndSignal;
                }
                break;
            }
        }

        // Event loop: update all systems: our graphics using render callback function,events,timing...
        while(CRM32Pro.Update(&keyevent)) 
        {
            switch(keyevent.type)
            {
            case SDL_KEYUP:
                switch(keyevent.key.keysym.sym)
                {
                case SDLK_ESCAPE:							
                    // Cancel match, start stopping procedure. Only works on local matches or on the server side
                    if((Mode == GAMEMODE_SERVER) || (Mode == GAMEMODE_LOCAL))
                    {
                        iMatchFSM = 3;
                        iGameFlag = GAME_ABORT;
                    }
                    break;
                case SDLK_F1:
                    bDebug = bDebug * (-1);
                    break;
                case SDLK_F2:
                    if(bDebug == 1) IImage->SavePNG("screenshot.png",CRM32Pro.screen);
                    break;     
                case SDLK_F3:
                    if(bDebug == 1) BreakablesKillAll();
                    break;
                case SDLK_F5:
                    if(bDebug == 1) { if(tPlayers[0].player != NULL) { tPlayers[0].player->AddPoints(300,0,tPlayers[0].player->GetXr(),tPlayers[0].player->GetYr());  } }
                    break;
                }
                break;

            case EVENT_LOGICWAIT:
                // Render graphics
                //RenderGraphics(1);					
                break;
            }
        }		  
    }

    // Free all the resources and return to the game loop logic
    CRM32Pro.SetRenderCallback(NULL);
    IScreenFX->RenderFX(CRM32Pro.screen,CRM32Pro.screen,FXSCREEN_GREY);
    IScreenFX->FadeToColor(0,0,0,1000);	
    MapFree();
    PlayersFree();
    return done;
}

// Propagate the match end signal to all clients connected
// It bypass the server callback as this signal is only sent by the server so we assure only one sending point!
void cGameField::SendMatchEndSignal(int bMatchFlag,Uint8 iGStatus)
{
    TBDN_Event.iCtrlID = bMatchFlag;
    TBDN_Event.event.move_x = iGStatus; // send here the gamestatus
    TBDN_Event.event.move_y = 0;
    TBDN_Event.event.Xr = 0;
    TBDN_Event.event.Yr = 0;
    TBDN_Event.event.bomb = 0;
    TBDN_Event.event.item = ITEM_NULL;
    TBDN_Event.event.points = 0;
    TBDN_Event.event.death = 0;
    INetwork->SendData((char *)&TBDN_Event,sizeof(TBDN_Event),0); // Bypass server callback function
    ILogSystem.Msg(LOG_NORMAL," [GameField Info] - SendMatchEndSignal(%d,%d)\n",bMatchFlag,iGStatus);
}

// Render graphics callback with different modes (start, match and end)
void cGameField::RenderGraphics(int m)
{
    int i,j,k,l;
    SDL_Rect rect;

    // Is the map ready?
    if(!bMapReady) return;

    // Blit background
    SDL_BlitSurface(sGameBg,NULL,CRM32Pro.screen,NULL);

    // Blit blocks
    for(i = 0; i < 15; i++)
    {
        for(j = 0; j < 11; j++)
        {
            if(GetItem(i,j) == ITEM_BLOCK)
            {
                rect.h = sBlock->h; rect.w = sBlock->w;
                rect.x = (i * rect.w) + (Sint16)ORIGIN_X; rect.y = (j * rect.h) + (Sint16)ORIGIN_Y;
                SDL_BlitSurface(sBlock,NULL,CRM32Pro.screen,&rect);
            }
        }  
    }

    // Blit breakables
    for(k = 0; k < tMapGame.iNumBK; k++)
    {
        // is there a breakable?
        if(tBreakables[k].iItem >= ITEM_BREAKABLE)
        {		
            switch(tBreakables[k].iStatus)
            {
                // Normal status: only draw first frame
            case 1: 
                tBreakables[k].sprBreakable->Draw();
                break;
                // Explosion status
            case 4:
                tBreakables[k].sprBreakable->Resume();
                tBreakables[k].sprBreakable->Draw();					 
                break;
                // It has an item
            case 5:
                tBreakables[k].sprItem->Draw();
                break;
            }
        }		 
    }

    // Blit Player and status
    for(i = 0; i < 4; i++)
    {
        // Is there a player?
        if(tPlayers[i].player != NULL) 
        {
            // Draw the player and bombs
            tPlayers[i].player->BombRender();
            tPlayers[i].player->Render();				

            // Life icon
            tIcons->SetPosition((OINFO_UNIT*i)+OINFO_X+ITEM1_X,OINFO_Y+ITEM1_Y);
            tIcons->Draw(CRM32Pro.screen,1);	  
            // Bomb icon
            tIcons->SetPosition((OINFO_UNIT*i)+OINFO_X+ITEM1_X,OINFO_Y+ITEM2_Y);	  
            tIcons->Draw(CRM32Pro.screen,2);
            // Power icon
            tIcons->SetPosition((OINFO_UNIT*i)+OINFO_X+ITEM2_X,OINFO_Y+ITEM1_Y);
            tIcons->Draw(CRM32Pro.screen,3);
            // Speed icon
            tIcons->SetPosition((OINFO_UNIT*i)+OINFO_X+ITEM2_X,OINFO_Y+ITEM2_Y);
            tIcons->Draw(CRM32Pro.screen,4);
            // Draw current player stats
            tPlayers[i].player->InfoPlayer();
        }
    }

    // Debug information 
    if(bDebug == 1)// && (ITimeSystem->GetTime()>(iDebugUpdate+100)))
    {
        iDebugUpdate = ITimeSystem->GetTime();
        for(i = 0; i < 4; i++)
        {		  
            if(tPlayers[i].player!=NULL) tPlayers[i].player->Debug();
        }
        Debug();
    } 


    // Loop through our list rendering or removing floating messages
    itFloatingMsg = vFloatingMsgs.begin();
    while(itFloatingMsg != vFloatingMsgs.end())
    {
        // Is alive this message?
        i = ITimeSystem->GetTime();
        if(i < (itFloatingMsg->iStart + itFloatingMsg->iTTL))
        {
            // with iTTL/2, we start to move up the message. Up to 15pixels
            if(i > (itFloatingMsg->iStart + (itFloatingMsg->iTTL / 2))) 
            {
                j = ((i - itFloatingMsg->iStart - (itFloatingMsg->iTTL / 2)) * (15*2) ) / itFloatingMsg->iTTL;
            }
            else j = 0;
            itFloatingMsg->fMsg->PutString(CRM32Pro.screen,itFloatingMsg->x,itFloatingMsg->y - j,(char*)itFloatingMsg->sPayload);
            ++itFloatingMsg;
        }
        // is dead, fading effect and we remove it
        else
        {
            //ILogSystem.Msg(LOG_NORMAL," > Removed floating message (%d,%d): '%s'\n",itFloatingMsg->x,itFloatingMsg->y,itFloatingMsg->sPayload);
            itFloatingMsg = vFloatingMsgs.erase(itFloatingMsg);
        }
    }

    // If we are on the MessageStartGame(), we draw the associated graphics
    if(iRenderGraphicsStatus == RENDERGRAPHICS_START)
    {		
        SpecialText(CRM32Pro.screen,FontStore->Info,CRM32Pro.screen->w/2 - FontStore->Info->StringWidth(sMsg)/2,200,sMsg,iSpecialTextCounter);
        FontStore->Info->PutString(CRM32Pro.screen,CRM32Pro.screen->w/2 - FontStore->Info->StringWidth(sTmp)/2,250,sTmp);	
    }
    // If we are on the MessageEndGame(), we draw the associated graphics
    else if(iRenderGraphicsStatus == RENDERGRAPHICS_END)
    {
        SDL_Rect rMyRect;
        rMyRect.x = 250; rMyRect.y = 180;
        string strTmp;

        // Render background
        CRM32Pro.Blit(sEndMatchBg,NULL,CRM32Pro.screen,&rMyRect);
        // Render the match winner if any
        if(strlen(szWinner) > 0) FontStore->Info->PutString(CRM32Pro.screen,CRM32Pro.screen->w/2 - FontStore->Info->StringWidth(szWinner)/2,200,szWinner);	
        SpecialText(CRM32Pro.screen,FontStore->Info,CRM32Pro.screen->w/2 - FontStore->Info->StringWidth(sMsg)/2,250,sMsg,iSpecialTextCounter);	
        // Render total score table
        k = CRM32Pro.screen->w / 2 - FontStore->Info->StringWidth("Total score") / 2;
        FontStore->Info->PutString(CRM32Pro.screen,k,310,"Total score");
        // l will have the maximum player name size
        // i is the y position on the screen. We also use it for displaying the position number
        i = 340; l = 0; k -= 10; 
        // Render the players names
        for(j = 0; j < 4; j++) 
        {
            if(strcmp(stScore[j].sName,""))
            {
                _itoa(((i - 340) / 25) + 1,sTmp,10);
                strTmp = sTmp;
                strTmp += ". ";
                strTmp += stScore[j].sName;
                if(l < FontStore->Info->StringWidth((char*)strTmp.c_str())) l = FontStore->Info->StringWidth((char*)strTmp.c_str());
                FontStore->Info->PutString(CRM32Pro.screen,k,i,(char*)strTmp.c_str());
                i += 25;
            }
        }			
        // l has the maximum player name size
        i = 340; k += l + 15; 
        l = 0;
        // Render the players points
        for(j = 0; j < 4; j++) 
        {
            if(strcmp(stScore[j].sName,""))
            {
                _itoa(stScore[j].iPoints,sTmp,10);
                if(l == 0) l = FontStore->Score->StringWidth(sTmp);  // Now, flag will have the maximum points size														
                FontStore->Score->PutString(CRM32Pro.screen,k + (l - FontStore->Score->StringWidth(sTmp)),i,sTmp);
                i += 25;
            }
        }	
    }
    // We are on the playing field. Graphics that are only rendered on this state
    else
    {
        // If we are on a timed match... the counterdown
        if(strlen(sCounterDown) > 0)
        {
            FontStore->Score->PutStringXCenter(CRM32Pro.screen,CRM32Pro.screen->h/2,sCounterDown);
        }
    }
}

// Add points to a player: share points if value has different players
void cGameField::SetPoints(Uint8 value,int points,int x, int y)
{
    int iTmp = 0;

    if(value & PLAYER_WATER) iTmp++;
    if(value & PLAYER_FIRE) iTmp++;
    if(value & PLAYER_EARTH) iTmp++;
    if(value & PLAYER_AIR) iTmp++;

    if(value & PLAYER_WATER) if(tPlayers[0].player != NULL) tPlayers[0].player->AddPoints(points / iTmp,0,x,y);
    if(value & PLAYER_FIRE) if(tPlayers[1].player != NULL) tPlayers[1].player->AddPoints(points / iTmp,0,x,y);
    if(value & PLAYER_EARTH) if(tPlayers[2].player != NULL) tPlayers[2].player->AddPoints(points / iTmp,0,x,y);
    if(value & PLAYER_AIR) if(tPlayers[3].player != NULL) tPlayers[3].player->AddPoints(points / iTmp,0,x,y);
}

// Return the string with the name(s) of the players given an ID
char *cGameField::GetPlayerName(int lookforID)
{
    int i,j = 0;
    static string strNames;
    strNames = "";

    for(i = 0; i < 4; i++)
    {
        if(tPlayers[i].id & lookforID) 
        {
            if(j !=0 ) strNames += " and ";
            j++;
            strNames += tPlayers[i].name;
        }
    }

    return (char*)strNames.c_str();
}

// Devuelve el modo de juego
Uint8 cGameField::GetMode()
{
    return Mode;
}

void cGameField::Debug()
{   
    int i,j;
    char sDebug[15];// Debug string 
    char sTmp[4];   // Auxiliar string
    strcpy(sDebug,"");
    strcpy(sTmp,"");

    // Comprueba que se haya iniciado el mapa
    if(!bMapReady) return;

    for(i = 0; i < MAP_WIDTH; i++)
        for(j = 0; j < MAP_HEIGHT; j++)
        {
            _itoa(i,sTmp,10);
            strcpy(sDebug,"(");
            strcat(sDebug,sTmp);
            strcat(sDebug,",");
            _itoa(j,sTmp,10);
            strcat(sDebug,sTmp);
            strcat(sDebug,")");
            FontStore->Debug->PutString(CRM32Pro.screen,10+(DELTA_X*i)+ORIGIN_X,(DELTA_Y*j)+ORIGIN_Y,sDebug);
            strcpy(sDebug,"S: ");
            _itoa(tMapGame.tMapBomb[i][j].status,sTmp,10);
            strcat(sDebug,sTmp);
            FontStore->Debug->PutString(CRM32Pro.screen,8+(DELTA_X*i)+ORIGIN_X,(DELTA_Y*j)+ORIGIN_Y+14,sDebug);
            strcpy(sDebug,"I: ");
            _itoa(GetItem(i,j),sTmp,10);
            strcat(sDebug,sTmp);
            FontStore->Debug->PutString(CRM32Pro.screen,8+(DELTA_X*i)+ORIGIN_X,(DELTA_Y*j)+ORIGIN_Y+28,sDebug);
        }
}
