/*----------------------------------------------------------------------
TitanBomber - Roberto Prieto
Copyright (C) 2005-2023 MegaStorm Systems
contact@megastormsystems.com - http://www.megastormsystems.com

This software is provided 'as-is', without any express or implied
warranty.  In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must not
claim that you wrote the original software. If you use this software
in a product, an acknowledgment in the product documentation would be
appreciated but is not required.
2. Altered source versions must be plainly marked as such, and must not be
misrepresented as being the original software.
3. This notice may not be removed or altered from any source distribution.

------------------------------------------------------------------------

gameField class

------------------------------------------------------------------------ */

#ifndef _GAMEFIELD
#define _GAMEFIELD

// Flags for debugging
#define DEBUG_CORESERVER
#define DEBUG_NETWORK
//#define DEBUG_NETWORK_FULL

#include "CRM32Pro.h"
#include "titanbomber.h"
#include <math.h>
#include <queue>  // STL queue
#include <vector> // STL vector
using namespace std;

#ifndef _MACOSX
#define M_PI 3.14159265358979323846
#endif

#define GAMEMODE_LOCAL   1
#define GAMEMODE_CLIENT  2
#define GAMEMODE_SERVER  4
#define GAMESERVER_PORT 22

#define ITEM_NULL          0 
#define ITEM_BLOCK         1
#define ITEM_BREAKABLE     2
#define ITEM_BK_BOMB       3
#define ITEM_BK_FIREUP     4
#define ITEM_BK_SPEEDUP    5
#define ITEM_BK_SPEEDDOWN  6

#define STATUS_NULL       0
#define STATUS_NOWALKABLE 1
#define STATUS_WALKABLE   2
#define STATUS_RESERVED1  4 // Reserved for players' control
#define STATUS_RESERVED2  8 // Reserved for players' control
#define STATUS_BURN_P1   TBDN_P1EVENT // 16    
#define STATUS_BURN_P2   TBDN_P2EVENT // 32
#define STATUS_BURN_P3   TBDN_P3EVENT // 64  
#define STATUS_BURN_P4   TBDN_P4EVENT //128   

#define MAX_BREAKABLES 100
#define ITEMS_BOMBS     10
#define ITEMS_FIREUP     8
#define ITEMS_SPEEDUP    8
#define ITEMS_SPEEDDOWN  6

#define GAME_GO    1
#define GAME_END   TBDN_GAMEEND   // 2
#define GAME_ABORT TBDN_GAMEABORT // 3

#define TIME_EOM	2000

#define POINTS_BOMB       10       // Points given by each bomb explosion
#define POINTS_BK        100       // Points given to break a breakable
#define POINTS_ITEM       50       // Points given to get/destroy an item
#define POINTS_KILLP    2000       // Points given to kill a human player
#define POINTS_AUTOKILL -500       // Points given to auto kill (rest points)
#define POINTS_KILLCPU  1000       // Points given to kill a cpu player
#define POINTS_LOCALWIN    120     // Percent to multiply the total score of local winner (fixed, =1.20) 
#define POINTS_VISITORWIN  150     // Percent to multiply the total score of visitor winner (fixed, =1.50)

#define PLAYER_WATER   TBDN_P1EVENT // 16   
#define PLAYER_FIRE    TBDN_P2EVENT // 32  
#define PLAYER_EARTH   TBDN_P3EVENT // 64  
#define PLAYER_AIR     TBDN_P4EVENT //128 

// Map vars
#define NUMBER_OF_MAP 4
#define MAP_HEIGHT 11
#define MAP_WIDTH 15

// Screen coordinates of each player information
// Start position of pannel
#define OINFO_X      2 
#define OINFO_Y      9
#define OINFO_UNIT 198
// Relative position of score
#define SCORE_X      5
#define SCORE_Y     23
// Relative position of items and fonts
#define ITEM1_X    107
#define FONT1_X    125
#define ITEM2_X    149
#define FONT2_X    167
#define ITEM1_Y      3
#define FONT1_Y      2
#define ITEM2_Y     24
#define FONT2_Y     23

#define RENDERGRAPHICS_START 0
#define RENDERGRAPHICS_GAME  1
#define RENDERGRAPHICS_END   2

// RenderInGame callback wrapper
void myRenderInGame(int i);

class cGameField
{
    friend void myRenderInGame(int);
public:
    cGameField();
    ~cGameField();
    Uint8 Init(int,char *,sPlayers *,int,int);  // Initialize a game	
    Uint8 Execute(); // Execution of loaded map
    Uint8 GetStatus(int,int);
    void Close(); // Close current match but leaves it ready to init another one
    void SetStatus(int,int,Uint8);
    Uint8 GetItem(int,int);	
    void AddFloatingMsg(int,int,char *,CRM32Pro_CFont *,int);
    void SetItem(int,int,Uint8);
    Uint8 BreakablesKill(int,int);
    char *GetMapName();        // Return current loaded map name
    char *GetMapName(int nMap);// Return map name of the given position
    char *GetPlayerName(int);  // Return the name of player ID 
    char *GetGameTypeName();   // Return the gametype name
    int GetGameType();
    int GetPlayedTime();	   // Return the played time of the previous game in ms
    Uint8 GetMode();           // Return game mode	

private:
    void BreakablesKillAll();
    void MapLoad(char *);  // Load a map
    void MapInit();        // After a map and init game, initialize the map
    void MapFree();        // Free all map resources
    void PlayersInit();    // Initialize struct with all players
    void PlayersFree();    // Free all players resources
    void Debug();          // Show debug information on game screen
    void BreakablesExecute();  // Draw the breakables blocks
    void SetPoints(Uint8,int,int,int); // Search for given player and add its points	
    int  MessageStartGame();   // Show a message at the beginning of the game
    void MessageEndGame(int);  // Show a message at the end of the game
    void SpecialText(SDL_Surface *,CRM32Pro_CFont *, int, int, char *,int); // Show a text message with a special effect
    void RenderGraphics(int); // Graphics render callback 
    int  InitServer(char *mapname);	 // Initialize the game server
    int  InitClient(char *server_ip);// Initialize the client server
    void SendMatchEndSignal(int bMatchFlag,Uint8 iGStatus);  // Propagate the match end signal to all clients connected
    static int myCoreServer(void *data,int size); // Authorization callback server

    // Variables
    CRM32Pro_CSprite *sprBreakableMaster;
    CRM32Pro_CSprite *sprItemMaster;
    struct                // Breakables array
    {
        Uint8 iStatus;
        Uint8 iItem;
        Uint8 iX,iY;
        CRM32Pro_CSprite *sprBreakable; // Breakable sprite cloned
        CRM32Pro_CSprite *sprItem;      // Item sprite cloned
    } tBreakables[MAX_BREAKABLES];

    Uint8 Mode; // Match mode: GAMEMODE_LOCAL, GAMEMODE_CLIENT or GAMEMODE_SERVER
    Uint8 Type;	// Match type: GAMETYPE_LEAGUE, GAMETYPE_LEAGUE_LAST, GAMETYPE_TIMED or GAMETYPE_RANDOM

    Uint32 bServer;
    char host[20];
    char name[20];	
    SDL_Surface *sGameBg, *sEndMatchBg;	
    CRM32Pro_CTile *tIcons;
    SDL_Surface *sBlock;
    char bMapReady;  // 0: if map is wrong and 1: if map was loaded successfully
    // Debug stuff
    char bDebug;
    Uint32 iDebugUpdate;
    FILE *fp1,*fp2,*fp3,*fp4;

    // Players number (gets it from tPlayers struct)
    Uint8 iNumPlayers;
    Uint8 iNumRemotePlayers;
    // Pointer to original tPlayers struct
    sPlayers *tPlayers;

    // Game map struct
    struct
    {
        char sName[32];        // Name ID
        char sBg[32];          // IPF name with background
        char sBlock[32];       // Sprite name of non-break blocks
        char sBreakable[32];   // Sprite name of breakable blocks
        unsigned char iNumBK;  // Number of breakables [0-99]
        unsigned char iOwner;  // Ownership (PLAYER_WATER..)
        struct 
        {
            Uint8 item;
            Uint8 status;			
        } tMapBomb[MAP_WIDTH][MAP_HEIGHT];
    } tMapGame,lstMapGame[NUMBER_OF_MAP];

    // Access to map mutex (only for server on network games)
    SDL_mutex *mutMap;

    // Send/reception map/players struct
    sTBDN TBDN;	
    // Send/reception event struct
    sTBDN_Event TBDN_Event;
    // Reception queue for each player in client/server game mode
    queue<sTBDN_Event> TBDN_Events[4];
    // Delay before to start the game. On each match, the server update this value
    Uint32 iTimeStartDelay;
    // End of game vars
    Uint32 iMatchFSM;
    int iTimeEOM;
    // Game field total played time
    int iTimeStart, iTimeEnd, iTimeMatch;
    char sCounterDown[8];
    float fFPS;
    // Special messages (start game, end game, add points, etc.) vars
    int iRenderGraphicsStatus;
    int iSpecialTextCounter;   
    char sTmp[8];
    char sMsg[32];
    char szWinner[8];
    struct stScore
    {
        char sName[8];
        int  iPoints;
    } stScore[4];
    struct stFloatingMsg
    {
        int x,y;
        CRM32Pro_CFont *fMsg;
        char sPayload[16];
        int iTTL,iStart;
    };
    vector<stFloatingMsg> vFloatingMsgs;
    vector<stFloatingMsg>::iterator itFloatingMsg;

public:
    sTBDN_Event *GetEventForPlayer(int);
};

// Pointer to instanced gamefield used by myCoreServer static function
extern cGameField *myGameField;
#endif
